"""
    eu_air(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve air traffic data for specified geographic regions within Europe from a MySQL database.

This function queries a database for air traffic information, such as the number of airports and total flights, within specified geographic regions in Europe. The query filters data based on the shapes and IDs provided in the `selection` dictionary. Acceptable years for the data are 1961, 71, 81, 91, 2001, 2011.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `selection::Dict`: A dictionary containing `geo_id` and `shape_obj`, representing geographic identifiers and their corresponding shapes.

# Returns
- `DataFrame`: A DataFrame with columns for the geographic ID, the number of airports, the number of flights, and the region set to 'EU'.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> selection = Dict("geo_id" => ["ID1", "ID2"], "shape_obj" => ["SHAPE1", "SHAPE2"])
> df = eu_air(conn, selection)

This example queries air traffic data for specified European regions and returns the results in a DataFrame.
"""
function eu_air(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary: acceptable years are 1961, 71, 81, 91, 2001, 2011
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape) in zip(geo_id, shape_obj)
        # prepare query
        query = "SELECT COUNT(DISTINCT(ea.airp_icao_)) as airports, SUM(et.flights) as flights FROM eu_airports_2013 ea JOIN eu_traffic et ON ea.airp_icao_ = et.ICAO_id WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), ea.SHAPE) AND et.Year = 2022;"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    us_air(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve air traffic data for specified geographic regions within the United States from a MySQL database.

This function queries a database for air traffic information, such as the number of airports and total flights, within specified geographic regions in the United States. The query filters data based on the shapes and IDs provided in the `selection` dictionary. Acceptable years for the data are 1961, 71, 81, 91, 2001, 2011.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `selection::Dict`: A dictionary containing `geo_id` and `shape_obj`, representing geographic identifiers and their corresponding shapes.

# Returns
- `DataFrame`: A DataFrame with columns for the geographic ID, the number of airports, the number of flights, and the region set to 'US'.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> selection = Dict("geo_id" => ["ID1", "ID2"], "shape_obj" => ["SHAPE1", "SHAPE2"])
> df = us_air(conn, selection)

This example queries air traffic data for specified U.S. regions and returns the results in a DataFrame.
"""
function us_air(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary: acceptable years are 1961, 71, 81, 91, 2001, 2011
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape) in zip(geo_id, shape_obj)
        # prepare query
        query = "SELECT count(*) as airports, SUM(COALESCE(COMMERCIAL, 0) + COALESCE(COMMUTER_O, 0) + COALESCE(AIR_TAXI_O, 0) + COALESCE(LOCAL_OPS, 0) + COALESCE(ITNRNT_OPS, 0) + COALESCE(MIL_ACFT_O, 0)) as flights FROM us_airports WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE);"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("US", nrow(dftemp))
    return dftemp
end

"""
    get_airports_traffic(table::String, selection::Dict; host::String="127.0.0.1") -> DataFrame

Retrieve airport traffic data, either from Europe or the United States, based on the specified table and selection criteria.

This function determines whether to fetch European or U.S. air traffic data based on the provided `table` argument. It then retrieves the relevant data using the appropriate function (`eu_air` or `us_air`) and the given `selection` dictionary.

# Arguments
- `table::String`: Specifies whether to retrieve data for Europe ('eu_air') or the United States ('us_air').
- `selection::Dict`: A dictionary containing selection criteria.
- `host::String` (optional): The hostname for the MySQL server, defaulting to '127.0.0.1'.

# Returns
- `DataFrame`: A DataFrame containing air traffic data, with columns for region, geographic ID, number of airports, and number of flights.

# Example
> selection = Dict("geo_id" => ["ID1"], "shape_obj" => ["SHAPE1"])
> df = get_airports_traffic("eu_air", selection)

In this example, European air traffic data for the specified region is retrieved and returned in a DataFrame.
"""
function get_airports_traffic(table::String, selection::Dict; host::String="127.0.0.1")
    conn_dict = Dict(
        "host" => host,
        "port" => 3310,
        "database" => "FACT_air",
        "user" => "root",
        "password" => "devops"
    )

    table_dict = Dict(
        "eu_air" => eu_air,
        "us_air" => us_air
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        airports = Int64[], # column for number of airports
        flights = Int64[] # column to store number of flights
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end

    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 
    
    return df
end

